from dataframe import dataframe as df


class OneHotEncoder:
    def __init__(self):
        pass

    def fit(self, X: df.DataFrame, Y: df.Column | list):
        pass

    def transform(self, X: df.DataFrame) -> df.DataFrame:
        pass


class MinMaxScaler:
    def __init__(self):
        pass

    def fit(self, X: df.DataFrame, Y: df.Column | list):
        pass

    def transform(self, X: df.DataFrame) -> df.DataFrame:
        pass


class StandardScaler:
    def __init__(self):
        pass

    def fit(self, X: df.DataFrame, Y: df.Column | list):
        pass

    def transform(self, X: df.DataFrame) -> df.DataFrame:
        pass